@@ -1,15 +1,21 @@
 # -*- test-case-name: foolscap.test.test_pb -*-
 
-import os.path, weakref
+import os.path, weakref, binascii
 from zope.interface import implements
-from twisted.internet import defer, protocol
+from twisted.internet import defer, protocol, error
 from twisted.application import service, strports
-from twisted.python import log
+from twisted.python.failure import Failure
 
-from foolscap import ipb, base32, negotiate, broker, observer, eventual
+from foolscap import ipb, base32, negotiate, broker, observer, eventual, storage
+from foolscap import util
 from foolscap.referenceable import SturdyRef
-from foolscap.tokens import PBError, BananaError
+from foolscap.tokens import PBError, BananaError, WrongTubIdError, \
+     WrongNameError, NoLocationError
 from foolscap.reconnector import Reconnector
+from foolscap.logging import log as flog
+from foolscap.logging import log
+from foolscap.logging import publish as flog_publish
+from foolscap.logging.log import WEIRD, UNUSUAL
 
 crypto_available = False
 try:
@@ -29,6 +35,8 @@ class Listener(protocol.ServerFactory):
     Unauthenticated Tubs use a TubID of 'None'. There may be at most one such
     Tub attached to any given Listener."""
 
+    noisy = False
+
     # this also serves as the ServerFactory
 
     def __init__(self, port, options={},
@@ -117,10 +125,20 @@ class Listener(protocol.ServerFactory):
     def removeRedirect(self, tubID):
         del self.redirects[tubID]
 
+    def startFactory(self):
+        log.msg("Starting factory %r" % self)
+        return protocol.ServerFactory.startFactory(self)
+    def stopFactory(self):
+        log.msg("Stopping factory %r" % self)
+        return protocol.ServerFactory.stopFactory(self)
+
+
     def buildProtocol(self, addr):
         """Return a Broker attached to me (as the service provider).
         """
-        proto = self.negotiationClass()
+        lp = log.msg("%s accepting connection from %s" % (self, addr),
+                     addr=(addr.host, addr.port))
+        proto = self.negotiationClass(logparent=lp)
         proto.initServer(self)
         proto.factory = self
         return proto
@@ -176,6 +194,10 @@ class Tub(service.MultiService):
     @ivar referenceToName: maps Referenceable to a name
     @ivar nameToReference: maps name to Referenceable
 
+    @type tubID: string
+    @ivar tubID: a global identifier for this Tub, possibly including
+                 authentication information, hash of SSL certificate
+
     """
     implements(ipb.ITub)
 
@@ -190,6 +212,7 @@ class Tub(service.MultiService):
     brokerClass = broker.Broker
     keepaliveTimeout = 4*60 # ping when connection has been idle this long
     disconnectTimeout = None # disconnect after this much idle time
+    tubID = None
 
     def __init__(self, certData=None, certFile=None, options={}):
         service.MultiService.__init__(self)
@@ -199,12 +222,17 @@ class Tub(service.MultiService):
         else:
             self.setupEncryption(certData)
 
+    def __repr__(self):
+        return "<Tub id=%s>" % self.tubID
+
     def setupEncryptionFile(self, certFile):
-        if os.path.exists(certFile):
+        try:
             certData = open(certFile, "rb").read()
-            self.setupEncryption(certData)
-        else:
-            self.setupEncryption(None)
+        except EnvironmentError:
+            certData = None
+        self.setupEncryption(certData)
+
+        if certData is None:
             f = open(certFile, "wb")
             f.write(self.getCertData())
             f.close()
@@ -221,11 +249,35 @@ class Tub(service.MultiService):
         self.myCertificate = cert
         self.tubID = crypto.digest32(cert.digest("sha1"))
 
+    def make_incarnation(self):
+        unique = os.urandom(8)
+        # TODO: it'd be nice to have a sequential component, so incarnations
+        # could be ordered, but it requires disk space
+        sequential = None
+        self.incarnation = (unique, sequential)
+        self.incarnation_string = binascii.b2a_hex(unique)
+
+    def getIncarnationString(self):
+        return self.incarnation_string
+
     def setup(self, options):
         self.options = options
+        self.logger = flog.theLogger
         self.listeners = []
         self.locationHints = []
 
+        # duplicate-connection management
+        self.make_incarnation()
+
+        # the master_table records the master-seqnum we used for the last
+        # established connection with the given tubid. It only contains
+        # entries for which we were the master.
+        self.master_table = {} # k:tubid, v:seqnum
+        # the slave_table records the (master-IR,master-seqnum) pair for the
+        # last established connection with the given tubid. It only contains
+        # entries for which we were the slave.
+        self.slave_table = {} # k:tubid, v:(master-IR,seqnum)
+
         # local Referenceables
         self.nameToReference = weakref.WeakValueDictionary()
         self.referenceToName = weakref.WeakKeyDictionary()
@@ -246,6 +298,15 @@ class Tub(service.MultiService):
 
         self._pending_getReferences = [] # list of (d, furl) pairs
 
+        self._logport = None
+        self._logport_furl = None
+        self._logport_furlfile = None
+
+        self._log_gatherer_furl = None
+        self._log_gatherer_furlfile = None
+
+        self._handle_old_duplicate_connections = False
+
     def setOption(self, name, value):
         if name == "logLocalFailures":
             # log (with log.err) any exceptions that occur during the
@@ -271,9 +332,105 @@ class Tub(service.MultiService):
             self.keepaliveTimeout = value
         elif name == "disconnectTimeout":
             self.disconnectTimeout = value
+        elif name == "logport-furlfile":
+            self.setLogPortFURLFile(value)
+        elif name == "log-gatherer-furl":
+            self.setLogGathererFURL(value)
+        elif name == "log-gatherer-furlfile":
+            self.setLogGathererFURLFile(value)
+        elif name == "bridge-twisted-logs":
+            assert value is not False, "cannot unbridge twisted logs"
+            if value is True:
+                return flog.bridgeLogsFromTwisted(self.tubID)
+            else:
+                # for tests, bridge logs from a specific twisted LogPublisher
+                return flog.bridgeLogsFromTwisted(self.tubID,
+                                                  twisted_logger=value)
+        elif name == "handle-old-duplicate-connections":
+            if value is True:
+                value = 60
+            self._handle_old_duplicate_connections = int(value)
         else:
             raise KeyError("unknown option name '%s'" % name)
 
+    def setLogGathererFURL(self, gatherer_furl):
+        assert not self._log_gatherer_furl and not self._log_gatherer_furlfile
+        self._log_gatherer_furl = gatherer_furl
+        self._maybeConnectToGatherer()
+
+    def setLogGathererFURLFile(self, gatherer_furlfile):
+        assert not self._log_gatherer_furl and not self._log_gatherer_furlfile
+        self._log_gatherer_furlfile = gatherer_furlfile
+        self._maybeConnectToGatherer()
+
+    def _maybeConnectToGatherer(self):
+        if not self.locationHints:
+            return
+        furls = []
+        if self._log_gatherer_furl:
+            furls.append(self._log_gatherer_furl)
+        if self._log_gatherer_furlfile:
+            try:
+                # allow multiple lines
+                for line in open(self._log_gatherer_furlfile, "r").readlines():
+                    furl = line.strip()
+                    if furl:
+                        furls.append(furl)
+            except EnvironmentError:
+                pass
+        for f in furls:
+            connector = self.connectTo(f, self._log_gatherer_connected)
+
+    def _log_gatherer_connected(self, rref):
+        # we want the logport's furl to be nailed down now, so we'll use the
+        # right (persistent) name even if the user never calls
+        # tub.getLogPortFURL() directly.
+        ignored = self.getLogPortFURL()
+        tubID = self.tubID
+        if tubID is None:
+            # RILogGatherer.logport requires a string for nodeid=
+            tubID = '<unauth>'
+        rref.callRemoteOnly('logport', tubID, self.getLogPort())
+
+
+    def getLogPort(self):
+        if not self.locationHints:
+            raise NoLocationError
+        return self._maybeCreateLogPort()
+
+    def _maybeCreateLogPort(self):
+        if not self._logport:
+            self._logport = flog_publish.LogPublisher(self.logger)
+        return self._logport
+
+    def setLogPortFURLFile(self, furlfile):
+        self._logport_furlfile = furlfile
+        self._maybeCreateLogPortFURLFile()
+
+    def _maybeCreateLogPortFURLFile(self):
+        if not self._logport_furlfile:
+            return
+        if not self.locationHints:
+            return
+        # getLogPortFURL() creates the logport-furlfile as a side-effect
+        ignored = self.getLogPortFURL()
+
+    def getLogPortFURL(self):
+        if not self.locationHints:
+            raise NoLocationError
+        if self._logport_furl:
+            return self._logport_furl
+        furlfile = self._logport_furlfile
+        # the Tub must be running and configured (setLocation) by now
+        self._logport_furl = self.registerReference(self.getLogPort(),
+                                                    furlFile=furlfile)
+        return self._logport_furl
+
+
+    def log(self, *args, **kwargs):
+        kwargs['tubID'] = self.tubID
+        return log.msg(*args, **kwargs)
+
     def createCertificate(self):
         # this is copied from test_sslverify.py
         dn = crypto.DistinguishedName(commonName="newpb_thingy")
@@ -312,7 +469,54 @@ class Tub(service.MultiService):
         if not self.encrypted and len(hints) > 1:
             raise PBError("Unauthenticated tubs may only have one "
                           "location hint")
+        if self.locationHints:
+            raise PBError("Tub.setLocation() can only be called once")
         self.locationHints = hints
+        self._maybeCreateLogPortFURLFile()
+        self._maybeConnectToGatherer()
+
+    def setLocationAutomatically(self, *extra_addresses):
+        """Determine one of this host's publically-visible IP addresses and
+        use it to set our location. This uses whatever source address would
+        be used to get to a well-known public host (A.ROOT-SERVERS.NET),
+        which is effectively the interface on which a default route lives.
+        This is neither very pretty (IP address instead of hostname) nor
+        guaranteed to work (it may very well be a 192.168 'private' address),
+        but for publically-visible hosts this will probably produce a useable
+        FURL.
+
+        This method returns a Deferred that will fire once the location is
+        actually established. Calls to registerReference() must be put off
+        until the location has been set. And of course, you must call
+        listenOn() before calling autoSetLocation()."""
+
+        # first, make sure the reactor is actually running, by using the
+        # eventual-send queue
+        d = eventual.fireEventually()
+
+        def _reactor_running(res):
+            assert self.running
+            # we can't use get_local_ip_for until the reactor is running
+            return util.get_local_ip_for()
+        d.addCallback(_reactor_running)
+
+        def _got_local_ip(local_address):
+            local_addresses = set(extra_addresses)
+            if local_address:
+                local_addresses.add(local_address)
+            local_addresses.add("127.0.0.1")
+            locations = set()
+            for l in self.getListeners():
+                portnum = l.getPortnum()
+                for addr in local_addresses:
+                    locations.add("%s:%d" % (addr, portnum))
+            locations = list(locations)
+            locations.sort()
+            assert len(locations) >= 1
+            location = ",".join(locations)
+            self.setLocation(location)
+        d.addCallback(_got_local_ip)
+        return d
 
     def listenOn(self, what, options={}):
         """Start listening for connections.
@@ -327,7 +531,7 @@ class Tub(service.MultiService):
         @return: The Listener object that was created. This can be used to
         stop listening later on, to have another Tub listen on the same port,
         and to figure out which port was allocated when you used a strports
-        specification of'tcp:0'. """
+        specification of 'tcp:0'. """
 
         if type(what) is str:
             l = Listener(what, options, self.negotiationClass)
@@ -361,6 +565,11 @@ class Tub(service.MultiService):
             t.listenOn(l)
         return t
 
+    def getTubID(self):
+        return self.tubID
+    def getShortTubID(self):
+        return self.tubID[:4]
+
     def connectorStarted(self, c):
         assert self.running
         self._activeConnectors.append(c)
@@ -373,8 +582,8 @@ class Tub(service.MultiService):
             # which never shut down a Tub aren't going to care), and it is
             # more important to let application code run normally than to
             # force an error here.
-            log.msg("Tub.connectorFinished: WEIRD, %s is not in %s"
-                    % (c, self._activeConnectors))
+            self.log("Tub.connectorFinished: WEIRD, %s is not in %s"
+                     % (c, self._activeConnectors), level=WEIRD)
             return
         self._activeConnectors.remove(c)
         if not self.running and not self._activeConnectors:
@@ -390,9 +599,9 @@ class Tub(service.MultiService):
         for rc in self.reconnectors:
             eventual.eventually(rc.startConnecting, self)
 
-    def _tubsAreNotRestartable(self):
+    def _tubsAreNotRestartable(self, *args, **kwargs):
         raise RuntimeError("Sorry, but Tubs cannot be restarted.")
-    def _tubHasBeenShutDown(self):
+    def _tubHasBeenShutDown(self, *args, **kwargs):
         raise RuntimeError("Sorry, but this Tub has been shut down.")
 
     def stopService(self):
@@ -423,10 +632,11 @@ class Tub(service.MultiService):
 
         if self.brokers or self.unauthenticatedBrokers:
             dl.append(self._allBrokersAreDisconnected.whenFired())
+        why = Failure(error.ConnectionDone("Tub.stopService was called"))
         for b in self.brokers.values():
-            b.shutdown()
+            b.shutdown(why, fireDisconnectWatchers=False)
         for b in self.unauthenticatedBrokers:
-            b.shutdown()
+            b.shutdown(why, fireDisconnectWatchers=False)
 
         return defer.DeferredList(dl)
 
@@ -442,7 +652,7 @@ class Tub(service.MultiService):
             return "pb://" + self.tubID + "@" + hints + "/" + name
         return "pbu://" + self.locationHints[0] + "/" + name
 
-    def registerReference(self, ref, name=None):
+    def registerReference(self, ref, name=None, furlFile=None):
         """Make a Referenceable available to the outside world. A URL is
         returned which can be used to access this object. This registration
         will remain in effect (and the Tub will retain a reference to the
@@ -454,6 +664,11 @@ class Tub(service.MultiService):
                      name. If not, a random (unguessable) string will be
                      used.
 
+        @param furlFile: if provided, get the name from this file (if
+                         it exists), and write the new FURL to this file.
+                         If 'name=' is also provided, it is used for the
+                         name, but the FURL is still written to this file.
+
         @rtype: string
         @return: the URL which points to this object. This URL can be passed
         to Tub.getReference() in any Tub on any host which can reach this
@@ -463,11 +678,37 @@ class Tub(service.MultiService):
         if not self.locationHints:
             raise RuntimeError("you must setLocation() before "
                                "you can registerReference()")
+        oldfurl = None
+        if furlFile:
+            try:
+                oldfurl = open(furlFile, "r").read().strip()
+            except EnvironmentError:
+                pass
+        if oldfurl:
+            sr = SturdyRef(oldfurl)
+            if name is None:
+                name = sr.name
+            if self.tubID != sr.tubID:
+                raise WrongTubIdError("I cannot keep using the old FURL from %s"
+                                      " because it does not have the same"
+                                      " TubID as I do (%s)" %
+                                      (furlFile, self.tubID))
+            if name != sr.name:
+                raise WrongNameError("I cannot keep using the old FURL from %s"
+                                     " because you called registerReference"
+                                     " with a new name (%s)" %
+                                     (furlFile, name))
         name = self._assignName(ref, name)
         assert name
         if ref not in self.strongReferences:
             self.strongReferences.append(ref)
-        return self.buildURL(name)
+        furl = self.buildURL(name)
+        if furlFile:
+            need_to_chmod = not os.path.exists(furlFile)
+            open(furlFile, "w").write(furl + "\n")
+            if need_to_chmod:
+                os.chmod(furlFile, 0600)
+        return furl
 
     # this is called by either registerReference or by
     # getOrCreateURLForReference
@@ -655,11 +896,26 @@ class Tub(service.MultiService):
         if self.running:
             rc.startConnecting(self)
         else:
-            log.msg("Tub.connectTo(%s) queued until Tub.startService called"
-                    % _sturdyOrURL)
+            self.log("Tub.connectTo(%s) queued until Tub.startService called"
+                     % _sturdyOrURL, level=UNUSUAL)
         self.reconnectors.append(rc)
         return rc
 
+    def serialize(self, obj):
+        b = broker.StorageBroker(None)
+        b.setTub(self)
+        d = storage.serialize(obj, banana=b)
+        return d
+
+    def unserialize(self, data):
+        b = broker.StorageBroker(None)
+        b.setTub(self)
+        d = storage.unserialize(data, banana=b)
+        assert isinstance(d, defer.Deferred)
+        return d
+
+    # beyond here are internal methods, not for use by application code
+
     # _removeReconnector is called by the Reconnector
     def _removeReconnector(self, rc):
         self.reconnectors.remove(rc)
@@ -693,7 +949,8 @@ class Tub(service.MultiService):
         t1.setPeer(t2); t2.setPeer(t1)
         n = negotiate.Negotiation()
         params = n.loopbackDecision()
-        b1,b2 = self.brokerClass(params), self.brokerClass(params)
+        b1,b2 = (self.brokerClass(tubref, params),
+                 self.brokerClass(tubref, params))
         # we treat b1 as "our" broker, and b2 as "theirs", and we pretend
         # that b2 has just connected to us. We keep track of b1, and b2 keeps
         # track of us.
@@ -725,6 +982,7 @@ class Tub(service.MultiService):
                 d.errback(why)
 
     def brokerAttached(self, tubref, broker, isClient):
+        assert self.running
         if not tubref:
             # this is an inbound connection from an unauthenticated Tub
             assert not isClient
@@ -744,19 +1002,24 @@ class Tub(service.MultiService):
             del self.tubConnectors[tubref]
 
         if tubref in self.brokers:
-            # oops, this shouldn't happen but it isn't fatal. Raise
-            # BananaError so the Negotiation will drop the connection
+            # this shouldn't happen: acceptDecision is supposed to drop any
+            # existing old connection first.
+            self.log("ERROR: unexpected duplicate connection from %s" % tubref)
             raise BananaError("unexpected duplicate connection")
         self.brokers[tubref] = broker
 
         # now inform everyone who's been waiting on it
         if tubref in self.waitingForBrokers:
-            waiting = self.waitingForBrokers[tubref]
+            for d in self.waitingForBrokers[tubref]:
+                eventual.eventually(d.callback, broker)
             del self.waitingForBrokers[tubref]
-            for d in waiting:
-                d.callback(broker)
 
     def brokerDetached(self, broker, why):
+        # a loopback connection will produce two Brokers that both use the
+        # same tubref. Both will shut down about the same time. Make sure
+        # this doesn't confuse us.
+        had_connections = (bool(self.brokers) or
+                           bool(self.unauthenticatedBrokers))
         # the Broker will have already severed all active references
         for tubref in self.brokers.keys():
             if self.brokers[tubref] is broker:
@@ -764,21 +1027,40 @@ class Tub(service.MultiService):
         if broker in self.unauthenticatedBrokers:
             self.unauthenticatedBrokers.remove(broker)
         # if the Tub has already shut down, we may need to notify observers
-        # who are waiting for all of our connections to finish shutting down
+        # who are waiting for all of our connections to finish shutting down.
+        # Only do this if we actually transitioned from having some
+        # connections to not having any connections.
+
         if (not self.running
+            and had_connections
             and not self.brokers
             and not self.unauthenticatedBrokers):
             self._allBrokersAreDisconnected.fire(self)
 
+    def debug_listBrokers(self):
+        # return a list of (tubref, inbound, outbound) tuples. The tubref
+        # tells you which broker this is, 'inbound' is a list of
+        # InboundDelivery objects (one per outstanding inbound message), and
+        # 'outbound' is a list of PendingRequest objects (one per message
+        # that's waiting on a remote broker to complete).
+        output = []
+        all_brokers = (self.brokers.items()
+                       + [("unauth",broker)
+                          for broker in self.unauthenticatedBrokers])
+        for tubref,broker in all_brokers:
+            inbound = broker.inboundDeliveryQueue[:]
+            outbound = [pr
+                        for (reqID, pr) in
+                        sorted(broker.waitingForAnswers.items()) ]
+            output.append( (str(tubref), inbound, outbound) )
+        output.sort(lambda x,y: cmp( (len(x[1]), len(x[2])),
+                                     (len(y[1]), len(y[2])) ))
+        return output
+
+
 class UnauthenticatedTub(Tub):
     encrypted = False
 
-    """
-    @type tubID: string
-    @ivar tubID: a global identifier for this Tub, possibly including
-                 authentication information, hash of SSL certificate
-    """
-
     def __init__(self, tubID=None, options={}):
         service.MultiService.__init__(self)
         self.setup(options)
@@ -786,6 +1068,10 @@ class UnauthenticatedTub(Tub):
         assert not tubID # not yet
         self.tubID = tubID
 
+    def getTubID(self):
+        return "<unauth>"
+    def getShortTubID(self):
+        return "<unauth>"
 
 def getRemoteURL_TCP(host, port, pathname, *interfaces):
     url = "pb://%s:%d/%s" % (host, port, pathname)
